/*{{{  read me*/
/* pcl escape sequences come in two flavours...
 *
 * ESC X   - X between 48 and 126
 *
 * eg ESC E = reset
 *
 *
 * ESC X y # z1 # z2 # z3 ... # Zn
 *
 * X in 33 - 47  (! - /)
 * y = (optional ?) group 96-126
 * # = optional value field - omitted => 0
 *                            might have + or -
 *                            can have decimal point
 *
 * z = param, 96-126
 * Z = param  64-94
 *
 * above sequence is really just a condensed form of
 * ESC X y # Z1
 * ESC X y # Z2
 * ESC X y # Z3
 * ...
 * ESC X y # Zn
 *
 * eg ESC & l 1 o 2 A
 * ESC & l 1 O =  (something)
 * ESC & l 2 A =  (something else)
 *
 *
 *
 * Some sequences mean interpret the following data in a certain way
 * eg raster graphics. This sequence can be in the middle of a string,
 * so we must read in the entire string before processing it.
 */
/*}}}*/

/*{{{  system dependent stuff - define INT32 and PERCENT_D32*/
/* oops - forgot that dos only has 16 bits ! */

#if -1 == 65535U
typedef long INT32;
#define PERCENT_D32 "%ld"
#else
typedef int INT32;
#define PERCENT_D32 "%d"
#endif
/*}}}*/

/*{{{  environment array elements*/
/* for ease of tabulation, environment is an array - offsets are as follows */
/* leave 0 to mean none */

enum env_offsets {
ENV_COPIES=1,
ENV_SOURCE,
ENV_PAGE_SIZE,
ENV_ORIENT,
ENV_LEFT,
ENV_RIGHT,
ENV_TOP,
ENV_BOTTOM,
ENV_PERF,
ENV_HMI,
ENV_VMI,
ENV_UNDERLINE,

ENV_RASTER_RES,
ENV_RASTER_MARGIN,

ENV_AREA,
ENV_WIDTH,
ENV_HEIGHT,

ENV_MACRO,

ENV_LINE,
ENV_WRAP,

ENV_CURRENT_FONT,  /* controlled by SI/SO */

ENV_FONT_ID,
ENV_CHAR,

ENV_PRIM_SYMBOL_SET,
ENV_PRIM_ID,
ENV_PRIM_SPACING,
ENV_PRIM_PITCH,
ENV_PRIM_HEIGHT,
ENV_PRIM_STYLE,
ENV_PRIM_WEIGHT,
ENV_PRIM_TYPEFACE,

ENV_SCND_SYMBOL_SET,
ENV_SCND_ID,
ENV_SCND_SPACING,
ENV_SCND_PITCH,
ENV_SCND_HEIGHT,
ENV_SCND_STYLE,
ENV_SCND_WEIGHT,
ENV_SCND_TYPEFACE,

ENVIRONMENT_SIZE  /* must be last */
};

typedef enum env_offsets env_offsets;


/* one of these values is put in current_font, so that we can get
 * to the font attributes using
 * environment[ENV_PRIM_* + environment[ENV_CURRENT_FONT]]
 */
 
#define PRIMARY_FONT 0
#define SECOND_FONT (ENV_SCND_SYMBOL_SET - ENV_PRIM_SYMBOL_SET)
/*}}}*/

/*{{{  escape-sequence properties*/
/* properties in escape table */

#define MARKS_PAGE      1   /* this sequence causes output to page  */
#define CLEARS_PAGE     2   /* eg throws a page                     */
#define EXPECTS_DATA    4   /* command expects 'param' data points  */
#define FLUSH_PAGE      8   /* flush a page before this command     */
#define PAGE_SETUP     16   /* a page-setup ie resets margins, etc  */

/* we take charge of tracking cursor */
#define NEEDS_CURSOR         32
#define INVALIDATES_CURSOR   64

/* and font changes - we only mark font as invalid if it is ( and we are using
 * primary, or sequence is ) and we are using secondary
 */
#define INVALIDATES_FONT    128

/* notes
 *
 * we only tell the device to set the cursor just before it is used,
 * hence needs_cursor and invalidates_cursor
 *
 * typically, a job is a sequence of page-setup commands followed by
 * a sequence of data. We signal the device driver at the end of
 * the setup sequence
 */
/*}}}*/

/*{{{  structures*/
/*{{{  typedef struct cursor_struct  - x and y*/
typedef struct cursor_struct
{	struct cursor_struct *next;   /* for push/pop */
	int valid;  /* has it moved since last set_cursor ? */
	INT32 x;
	INT32 y;
} cursor_struct;
/*}}}*/

typedef struct escape_struct escape_struct;
typedef struct escape_entry_struct escape_entry_struct;
typedef void (*escape_handler_type)(escape_struct *escape);

/*{{{  struct escape_entry_struct*/
struct escape_entry_struct
{
	char command[4];
	char *desc;    /* for verbose mode - %d for (scaled) param */
	escape_handler_type action;
	env_offsets environment;  /* which environment element */
	int properties;
	int scaling_factor;
};
/*}}}*/

/*{{{  struct escape_struct*/
struct escape_struct
{
	/* briefly, either esc x  or esc x y # z - details above
	 * This is passed to all escape_handlers, if they need to know
	 * any details
	 */

	char string[4];  /* string[0] is x, [1] is y (' ' if omitted), [2] is z */

	char  *pointer;  /* in case we want to look for +/-, say */
	double real;    /* param as a real number */
	INT32  param;  /* param, scaled to centi-points if necessary */

	unsigned char *data;  /* for sequences that take data -  pcl.c reads it , */
	            /* in case device chooses not to implement/read it  */

	escape_entry_struct *entry;  /* table entry */
	            
};
/*}}}*/

/*}}}*/

/*{{{  variables and macros*/
/* ITS_PCL_C is defined if we are included from pcl.c
 * thus we define our vars, else just declare them
 */

#ifdef ITS_PCL_C
# define EXTERN
# define INIT(x) = x
#else
# define EXTERN extern
# define INIT(x)
#endif

EXTERN FILE *outfile;

EXTERN int *environment INIT(NULL);
EXTERN cursor_struct *cursor INIT(NULL);

EXTERN int debug INIT(0);
EXTERN int quiet INIT(0);

EXTERN int virgin_page  INIT(1);   /* is this a clean page ?              */
EXTERN int font_valid   INIT(0); /* has the device got the correct font */


/* these will be invoked with _two_ (('s - eg DEBUG(("x is %d", x)) */

#define DEBUG(x) do { if (debug ) to_stderr x; } while(0)
#define WARN(x)  do { if (!quiet) to_stderr x; } while(0)
/*}}}*/

/*{{{  scaling factors*/
/* scaling factors for various sizes */

#define SCALE_DECIPOINT 10
#define SCALE_DOT       24
#define SCALE_POINT    100
#define SCALE_48       150   /* 1/48 inch              */
#define SCALE_120       60  /* 1/120 inch             */
#define SCALE_ROWS     -1  /* (environment[ENV_HMI]) */
#define SCALE_COLS     -2 /* (environment[ENV_VMI]) */
/*}}}*/

/*{{{  paper sizes*/
#define UNPRINTABLE_TOP    (60*SCALE_DOT)
#define UNPRINTABLE_LEFT   (50*SCALE_DOT)
#define UNPRINTABLE_RIGHT  (100*SCALE_DOT)  /* approx */
#define UNPRINTABLE_BOTTOM (60*SCALE_DOT)  /* approx */

/* INIT doesn't work with structures... */
#ifdef ITS_PCL_C
INT32 paper_width[4] = { 2480*SCALE_DOT, 2550*SCALE_DOT, 2550*SCALE_DOT, 2175*SCALE_DOT };
INT32 paper_height[4]= { 3507*SCALE_DOT, 3300*SCALE_DOT, 4200*SCALE_DOT, 3150*SCALE_DOT };
#else
extern INT32 paper_width[4];
extern INT32 paper_height[4];
#endif

#define PAPER_WIDTH  ( paper_width[environment[ENV_PAGE_SIZE]])
#define PAPER_HEIGHT ( paper_height[environment[ENV_PAGE_SIZE]])
/*}}}*/

/*{{{  support routines in pcl.c*/
/* support routines in pcl.c */

void fatal(char *fmt, ...);
void to_stderr(char *fmt, ...);

void *xmalloc(int size);
void *xrealloc(void *p, int size);

void pcl_area_fill(escape_struct *escape);
/*}}}*/

/*{{{  fns provided by device-specific file*/
/* fns provided by device-specific file */

/* called once */
void dev_init(void);

/* pcl.c will call dev_throw_page if necessary before dev_close */
void dev_close(void);

/* guess */
void dev_reset(escape_struct *escape);

/* pcl.c keeps track of whether a page has been used in a variable
 * 'virgin_page' - will only call this if page is marked...
 */
void dev_throw_page(void);

/* called just before first mark on page */
void dev_about_to_mark_page();

/* called when element 'element' of the environment is changed */
void dev_env_changed(env_offsets element);

/* pcl.c tracks cursor, and calls this just before a write.
 * Position is in 'cursor' struct
 */
void dev_set_cursor(escape_struct *escape);

/* ditto for fonts - offset is 0 for primary, and offset from primary to
 * secondary environment data for secondary
 * variable 'font_valid' tracks changes to font characteristics since last
 * call to dev_set_font
 */
void dev_set_font(int font_offset);

/* routines for raster graphics
 * pcl.c keeps the cursor co-ords up to date, but it's up to the
 * device to call dev_set_cursor to actually position it, if it
 * wants to - ie, cursor->x & y are moved, but cursor->valid is not
 * set false
 */
void dev_begin_raster(escape_struct *escape);
void dev_raster_data(escape_struct *escape);
void dev_end_raster(escape_struct *escape);

/* Details of area are in the environment.
 * May call pcl_area_fill, which will deliver the area as a raster image
 * (not implemented yet, of course !)
 */
void dev_area_fill(escape_struct *escape);

/* soft-font stuff - probably not implemented */
void dev_font_control(escape_struct *escape);
void dev_select_font(escape_struct *escape);
void dev_font_descriptor(escape_struct *escape);
void dev_char_descriptor(escape_struct *escape);

/* called once before first char of a group (after cursor is set),
 * in case driver cares - eg postscript outputs a ( */
void dev_start_string(void);

/* write one character at current cursor posn
 * may set cursor->valid false if it wants pcl.c to dev_set_cursor()
 * before n